import unittest
import numpy as np
import time
import asyncio

from joule.models import DataStream, Element, pipes
from joule.models.data_store.nilmdb import NilmdbStore
from joule.models.data_store.errors import DataError
from tests.models.data_store.fake_nilmdb import FakeNilmdb
from tests import helpers
from tests.models.pipes.reader import QueueReader


class TestNilmdbInserter(unittest.IsolatedAsyncioTestCase):
    use_default_loop = False
    forbid_get_event_loop = True

    async def asyncSetUp(self):
        self.fake_nilmdb = FakeNilmdb()
        url = await self.fake_nilmdb.start()
        # use a 0 insert period for test execution
        self.store = NilmdbStore(url, 0, 60)
        await self.store.initialize([])
        # make a couple example streams
        # stream1 int8_3
        self.stream1 = DataStream(id=1, name="stream1", datatype=DataStream.DATATYPE.INT8,
                                  elements=[Element(name="e%d" % x) for x in range(3)])

    async def asyncTearDown(self):
        await self.fake_nilmdb.stop()
        await self.store.close()

    async def test_decimating_inserter(self):
        self.stream1.decimate = True
        source = QueueReader()
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        nrows = 955
        data = helpers.create_data(layout="int8_3", length=nrows)
        task = await self.store.spawn_inserter(self.stream1, pipe)
        for chunk in helpers.to_chunks(data, 300):
            await source.put(chunk.tobytes())
        await task
        self.assertEqual(len(self.fake_nilmdb.streams), 6)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].rows, nrows)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-4"].rows,
                         np.floor(nrows / 4))
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-16"].rows,
                         np.floor(nrows / 16))
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-64"].rows,
                         np.floor(nrows / 64))
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-256"].rows,
                         np.floor(nrows / 256))

    async def test_nondecimating_inserter(self):
        self.stream1.decimate = False
        self.stream1.datatype = DataStream.DATATYPE.UINT16
        source = QueueReader()
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        nrows = 896
        data = helpers.create_data(layout="uint16_3", length=nrows)
        task = await self.store.spawn_inserter(self.stream1, pipe, insert_period=0)
        for chunk in helpers.to_chunks(data, 300):
            await source.put(chunk.tobytes())
        await task

        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].rows, nrows)
        self.assertEqual(len(self.fake_nilmdb.streams), 1)

    async def test_propogates_intervals_to_decimations(self):
        self.stream1.decimate = True
        source = QueueReader()
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        # insert the following broken chunks of data
        # |----15*8----|-15-|-15-|-15-|-15-|  ==> raw (120 samples)
        # |-----30-----|-3--|--3-|--3-|--3-|  ==> x4  (27 samples)
        # |-----7------|                      ==> x16 (7 samples)
        # |-----1------|                      ==> x64 (1 sample
        n_chunks = 12
        for i in range(n_chunks):
            data = helpers.create_data(layout="int8_3", length=15, start=i * 1e6, step=1)
            await source.put(data.tobytes())
            if i > 6:  # breaks in the 2nd half
                await source.put(pipes.interval_token("int8_3").tobytes())
        task = await  self.store.spawn_inserter(self.stream1, pipe, insert_period=0)
        await task
        # should have raw, x4, x16, x64, x256
        self.assertEqual(len(self.fake_nilmdb.streams), 5)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1"].rows, n_chunks * 15)
        # x4 level should be missing data due to interval breaks
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-4"].rows, 42)
        # x16 level should have 7 sample (only from first part)
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-16"].rows, 7)
        # x64 level should have 1 sample
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-64"].rows, 1)
        # x256 level should be empty
        self.assertEqual(self.fake_nilmdb.streams["/joule/1~decim-256"].rows, 0)

    async def test_inserter_data_error(self):
        # when stream has invalid data (eg bad timestamps)
        self.stream1.datatype = DataStream.DATATYPE.UINT16
        source = QueueReader()
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        task = await self.store.spawn_inserter(self.stream1, pipe, insert_period=0)
        await source.put(helpers.create_data(layout="uint16_3").tobytes())
        # when nilmdb server generates an error
        self.fake_nilmdb.generate_error_on_path("/joule/1", 400, "bad data")
        with self.assertRaises(DataError):
            await task

    async def test_retries_when_nilmdb_is_not_available(self):
        # when nilmdb server is not available the inserter should retry
        self.stream1.datatype = DataStream.DATATYPE.UINT16
        source = QueueReader()
        await self.fake_nilmdb.stop()
        await source.put(helpers.create_data(layout="uint16_3").tobytes())
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        with self.assertLogs(level="WARNING") as logs:
            task = await self.store.spawn_inserter(self.stream1, pipe, retry_interval=0.05)
            await asyncio.sleep(0.1)
            task.cancel()
            await task
        self.assertTrue("retrying request" in ''.join(logs.output))

    async def test_inserter_clean(self):
        self.stream1.datatype = DataStream.DATATYPE.UINT16
        self.stream1.keep_us = 24 * 60 * 60 * 1e6  # 1 day
        self.stream1.decimate = True

        source = QueueReader(delay=0.1)
        await source.put(helpers.create_data(layout="uint16_3").tobytes())
        pipe = pipes.InputPipe(stream=self.stream1, reader=source)
        self.store.cleanup_period = 0
        task = await self.store.spawn_inserter(self.stream1, pipe, insert_period=0)
        await task

        self.assertTrue(len(self.fake_nilmdb.remove_calls) > 0)
        # make sure decimations have been removed too
        removed_paths = [x['path'] for x in self.fake_nilmdb.remove_calls]
        self.assertTrue('/joule/1' in removed_paths)
        self.assertTrue('/joule/1~decim-4' in removed_paths)
        self.assertTrue('/joule/1~decim-16' in removed_paths)
        # make sure nilmdb cleanup executed with correct parameters
        params = self.fake_nilmdb.remove_calls[-1]
        self.assertEqual(int(params['start']), 0)
        expected = int(time.time() * 1e6) - self.stream1.keep_us
        actual = int(params['end'])
        # should be within 0.1 second
        np.testing.assert_almost_equal(expected / 1e6, actual / 1e6, decimal=1)



from typing import Dict
from aiohttp import web
from aiohttp.test_utils import unused_port
import numpy as np
import multiprocessing
from tests import helpers


class FakeStream:
    def __init__(self, layout, start=None, end=None, rows=0, intervals=None):
        self.layout = layout
        self.start = start
        self.end = end
        self.rows = rows
        self.intervals = intervals
        self.metadata = {}
        if start is not None and end is not None:
            if self.intervals is None:
                self.intervals = [[start, end]]

    @property
    def dtype(self):
        ltype = self.layout.split('_')[0]
        lcount = int(self.layout.split('_')[1])
        atype = '??'
        if ltype.startswith('int'):
            atype = '<i' + str(int(ltype[3:]) // 8)
        elif ltype.startswith('uint'):
            atype = '<u' + str(int(ltype[4:]) // 8)
        elif ltype.startswith('float'):
            atype = '<f' + str(int(ltype[5:]) // 8)
        return np.dtype([('timestamp', '<i8'), ('data', atype, lcount)])


class FakeNilmdb:

    def __init__(self):
        self.app = web.Application()
        self.app.router.add_routes(
            [web.get('/', self.info),
             web.get('/dbinfo', self.dbinfo),
             web.get('/stream/get_metadata', self.stream_get_metadata),
             web.post('/stream/set_metadata', self.stream_set_metadata),
             web.post('/stream/create', self.stream_create),
             web.put('/stream/insert', self.stream_insert),
             web.get('/stream/extract', self.stream_extract),
             web.get('/stream/intervals', self.stream_intervals),
             web.post('/stream/remove', self.stream_remove),
             web.get('/stream/list', self.stream_list),
             web.post('/stream/destroy', self.stream_destroy)])
        self.runner = None
        self.error_on_paths = {}  # (msg, status) tuples to return for path
        self.stub_stream_create = False
        self.stub_stream_intervals = False
        self.http_code = 200
        self.response = ''
        # record of transactions
        self.posts = []
        self.gets = []
        self.extract_calls = []
        self.remove_calls = []
        self.destroy_calls = []
        self.stream_list_response = None
        # dict path=>layout from stream/create calls
        self.streams: Dict[str, FakeStream] = {}
        # optional queue of messages
        self.msgs: multiprocessing.Queue = None

    async def start(self, port=None, msgs: multiprocessing.Queue=None):
        if port is None:
            port = unused_port()
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, '127.0.0.1', port)
        await site.start()
        self.msgs = msgs
        return "http://localhost:%d" % port

    async def stop(self):
        await self.runner.cleanup()

    def generate_error_on_path(self, path, status, msg):
        self.error_on_paths[path] = (msg, status)

    async def info(self, request: web.Request):
        return web.Response(text="This is NilmDB version 1.11.0, running on host unittest")

    async def dbinfo(self, request: web.Request):
        return web.json_response(data={"path": "/opt/data",
                                       "other": 36002418688,
                                       "reserved": 11639611392,
                                       "free": 178200829952,
                                       "size": 2832261120})

    async def stream_create(self, request: web.Request):
        self.posts.append(request)
        if self.stub_stream_create:
            return web.Response(status=self.http_code, text=self.response)
        content = await request.post()

        if self.msgs is not None:
            self.msgs.put({'action': 'stream_create',
                           'params': dict(content)})

        if content["path"] in self.streams:
            return web.json_response({
                "status": "400 Bad Request",
                "message": "stream already exists at this path",
                "traceback": ""
            }, status=400)
        self.streams[content["path"]] = FakeStream(layout=content["layout"])
        return web.json_response(None)

    async def stream_insert(self, request: web.Request):
        self.posts.append(request)
        content = request.query
        if content['path'] in self.error_on_paths:
            (msg, status) = self.error_on_paths[content['path']]
            return web.Response(status=status, text=msg)
        stream = self.streams[content["path"]]
        start = int(content["start"])
        end = int(content["end"])

        # make sure the data bounds are valid
        if stream.rows > 0:
            if ((stream.start < start < stream.end) or
                    (stream.start < end < stream.end)):
                return web.json_response({
                    "status": "400 Bad Request",
                    "message": "new data overlaps existing range XXX",
                    "traceback": ""
                }, status=400)
        # update the stored bounds
        if stream.start is None:
            stream.start = start
        elif start < stream.start:
            stream.start = start
        if stream.end is None:
            stream.end = end
        elif end > stream.end:
            stream.end = end
        # update the data rows
        raw = await request.read()
        data = np.frombuffer(raw, stream.dtype)
        stream.rows += len(data)
        if self.msgs is not None:
            self.msgs.put({'action': 'stream_insert',
                           'params': dict(content), 'data': data})
        return web.json_response(None)

    async def stream_extract(self, request: web.Request):
        self.extract_calls.append(request.query)
        try:
            stream = self.streams[request.query["path"]]
        except KeyError:
            return web.json_response({
                "status": "404 Not Found",
                "message": "No such stream: %s" % request.query["path"],
                "traceback": ""
            }, status=404)

        if "count" in request.query:
            return web.Response(text=str(stream.rows))
        # return chunked data
        data = helpers.create_data(stream.layout, length=stream.rows)
        resp = web.StreamResponse()
        resp.enable_chunked_encoding()
        await resp.prepare(request)
        for chunk in helpers.to_chunks(data, 300):
            await resp.write(chunk.tobytes())
        return resp

    async def stream_intervals(self, request: web.Request):
        if self.stub_stream_intervals:
            return web.Response(status=self.http_code, text=self.response)
        try:
            stream = self.streams[request.query["path"]]
        except KeyError:
            return web.json_response({
                "status": "404 Not Found",
                "message": "No such stream: /joule/1",
                "traceback": ""
            }, status=404)
        if stream.intervals is None:
            return web.Response(status=200)

        text = "\r\n".join(["[%d, %d]" % (interval[0], interval[1]) for interval in stream.intervals])
        return web.Response(text=text)

    async def stream_remove(self, request: web.Request):
        self.remove_calls.append(request.query)
        return web.Response(text="100")

    async def stream_destroy(self, request: web.Request):
        content = await request.post()
        self.destroy_calls.append(content)
        return web.json_response(data=None)

    async def stream_list(self, request: web.Request):
        if self.stream_list_response is not None:
            return web.Response(text=self.stream_list_response)
        return web.json_response([[path,
                                   stream.layout,
                                   stream.start,
                                   stream.end,
                                   stream.rows,
                                   stream.end - stream.start] for path, stream in self.streams.items()])

    async def stream_get_metadata(self, request: web.Request):
        try:
            stream = self.streams[request.query["path"]]
            return web.json_response(stream.metadata)
        except KeyError:
            return web.Response(status=404)

    async def stream_set_metadata(self, request: web.Request):
        content = await request.post()
        if self.msgs is not None:
            self.msgs.put({'action': 'set_metadata',
                           'params': dict(content)})
        try:
            stream = self.streams[content["path"]]
            stream.metadata = dict(content)
        except KeyError:
            return web.Response(status=404)
        return web.Response(text="ok")



import psutil
import time

pid = 27557

while True:
    now = time.time()
    proc = psutil.Process(pid)
    memory = proc.memory_percent()
    with open('joule_stats.dat', 'a') as f:
        f.write("%d, %f\n" % (now, memory))
    time.sleep(1)


class MockSession:

    def __init__(self):
        self.path = None,
        self.response_data = None
        self.request_data = None
        self.method = None

    async def put(self, path, json):
        self.path = path
        self.request_data = json
        self.method = 'PUT'
        return self.response_data

    async def delete(self, path, params):
        self.path = path
        self.request_data = params
        self.method = 'DELETE'
        return self.response_data

    async def get(self, path, params):
        self.path = path
        self.request_data = params
        self.method = 'GET'
        return self.response_data

    async def post(self, path, json):
        self.path = path
        self.request_data = json
        self.method = 'POST'
        return self.response_data



#!/usr/bin/python3

"""
Run through each scenario in scenarios
"""
import os
import socket
import sys
import subprocess
import shlex
import shutil
import signal
import tempfile
import asyncio
import configparser
import time
import argparse

from joule import api

SOURCE_DIR = "/joule"
SCENARIO_DIR = "/joule/tests/e2e/scenarios"
MODULE_SCRIPT_DIR = "/joule/tests/e2e/module_scripts"
JOULE_CONF_DIR = "/etc/joule"
SECURITY_DIR = "/joule/tests/e2e/pki"

FORCE_DUMP = False


def prep_system():
    os.symlink(MODULE_SCRIPT_DIR, "/module_scripts")


def run(cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL):
    return subprocess.run(shlex.split(cmd), stdout=stdout, stderr=stderr)


async def wait_for_follower():
    # wait until the local node is online
    max_tries = 10
    num_tries = 0
    while num_tries < max_tries:
        num_tries += 1
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        result = sock.connect_ex(('localhost', 8088))
        sock.close()
        if result == 0:
            break
        time.sleep(0.5)
    if num_tries == max_tries:
        print("error, local node is not online")
        return -1

    node1 = api.get_node("node1.joule")
    while True:
        followers = await node1.follower_list()
        if len(followers) == 1:
            break
        await asyncio.sleep(0.5)
    await node1.close()
    return 0 # success

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--nilmdb", help="use nilmdb backend", action="store_true")
    args = parser.parse_args()

    prep_system()
    tests = []
    for entry in os.scandir(SCENARIO_DIR):
        if not (entry.name.startswith('.') and entry.is_dir()):
            if entry.name in ["__pycache__", "__init__.py"]:
                continue
            tests.append((entry.name, entry.path))

    # override to run only particular test(s)
    # tests = [("multinodes", "/joule/tests/e2e/scenarios/multiple_node_tests")]
    # tests = [("standalone modules", "/joule/tests/e2e/scenarios/standalone_modules")]
    # tests = [("API", "/joule/tests/e2e/scenarios/api_tests")]

    first_test = True
    for (test_name, test_path) in tests:
        if os.path.exists("/etc/joule/"):
            if os.path.islink("/etc/joule"):
                os.unlink("/etc/joule")  # this is a symlink
            else:
                shutil.rmtree("/etc/joule/")
        os.symlink(test_path, "/etc/joule")
        try:
            os.unlink("/etc/joule/security")
        except FileNotFoundError:
            pass
        os.symlink(SECURITY_DIR, "/etc/joule/security")


        if first_test:
            # get API permissions
            os.environ["LOGNAME"] = "e2e"
            os.environ["JOULE_USER_CONFIG_DIR"] = "/tmp/joule_user"
            with open(os.devnull, 'w') as devnull:
                subprocess.run("joule admin authorize".split(" "),
                               stdout=devnull)
            shutil.copy("/etc/joule/security/ca.joule.crt", "/tmp/joule_user/ca.crt")

        main_conf = "/etc/joule/main.conf"
        if args.nilmdb:
            # parse the main.conf file and enable nilmdb
            configs = configparser.ConfigParser()
            configs.read("/etc/joule/main.conf")
            configs["Main"]["NilmdbUrl"] = "http://nilmdb"
            (f, path) = tempfile.mkstemp()
            configs.write(os.fdopen(f, mode='w'))
            main_conf = path
        if first_test:
            jouled = subprocess.Popen(["jouled", "--config",
                                       main_conf],
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.STDOUT,
                                      universal_newlines=True)

            result = asyncio.run(wait_for_follower())
            jouled.send_signal(signal.SIGINT)
            stdout, _ = jouled.communicate()
            if result != 0:
                return result

        # clear the existing database (keeping master/follower tables)
        subprocess.run("joule admin erase --yes".split(" "))
        subprocess.run("joule admin authorize".split(" "))

        jouled = subprocess.Popen(["jouled", "--config",
                                   main_conf],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  universal_newlines=True)
        result = asyncio.run(wait_for_follower())

        print("---------[%s]---------" % test_name)
        sys.stdout.flush()

        test = subprocess.run(os.path.join(test_path, "test.py"))
        jouled.send_signal(signal.SIGINT)
        stdout, _ = jouled.communicate()
        if test.returncode != 0 or FORCE_DUMP:
            print("----dump from jouled----")
            for line in stdout.rstrip().split('\n'):
                print("> %s" % line)
            if test.returncode != 0:
                return test.returncode
    return 0  # success


if __name__ == "__main__":
    exit(main())



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from joule.models import master, Master
import joule.controllers
from tests.controllers.helpers import create_db, MockStore


class TestMasterController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app["db"].close()
        self.app["psql"].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0

        app["db"], app["psql"] = create_db([])
        app["data-store"] = MockStore()
        app["name"] = "test"
        app["port"] = 443
        app["scheme"] = "http"
        app["base_uri"] = "/"

        # create a master user to grant access
        self.grantor = Master(name="grantor", key="grantor_key",
                              type=Master.TYPE.USER)
        app["db"].add(self.grantor)
        app["db"].commit()
        return app


    async def test_adds_master_user(self):
        db: Session = self.app["db"]
        # add user to johndoe
        payload = {
            "master_type": "user",
            "identifier": "johndoe"
        }
        resp = await self.client.post("/master.json", json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertEqual(m.grantor_id, self.grantor.id)
        self.assertEqual(json['key'], m.key)

        # adding johndoe again returns an error and does not change the key
        resp = await self.client.post("/master.json", json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        old_key = m.key
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        new_key = m.key
        self.assertEqual(old_key, new_key)
        self.assertEqual(resp.status, 400)


    async def test_removes_master_user(self):
        db: Session = self.app["db"]
        db.add(Master(name="johndoe", key=master.make_key(), type=Master.TYPE.USER))
        db.commit()
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertIsNotNone(m)

        payload = {
            "master_type": "user",
            "name": "johndoe"
        }
        resp = await self.client.delete("/master.json", params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        m = db.query(Master).filter(Master.name == "johndoe").one_or_none()
        self.assertIsNone(m)


    async def test_lists_masters(self):
        db: Session = self.app["db"]
        db.add(Master(name="johndoe", key=master.make_key(), type=Master.TYPE.USER))
        db.add(Master(name="joule", key=master.make_key(), type=Master.TYPE.JOULE_NODE))
        db.add(Master(name="lumen", key=master.make_key(), type=Master.TYPE.LUMEN_NODE))
        db.commit()

        resp = await self.client.get("/masters.json",
                                     headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 200)
        json = await resp.json()
        # grantor is the original user
        self.assertEqual(len(json), 4)
        for item in json:
            self.assertTrue("key" not in item)
            self.assertTrue(item["name"] in ["johndoe", "joule", "lumen", "grantor"])
